#include "pwm.h"
#include <string.h>
#include <avr/io.h>

typedef struct {
	uint8_t port1;
	uint16_t nextTime;
} pwm_timer_t;

typedef enum {
	FALSE = 0,
	TRUE = 1,
} bool_t;

#define PWM_START 0
#define INF	0xffff;

static uint16_t dutyCycle[PWM_CHANNELS];				// 12 bit pwm values
static pwm_timer_t buffer[3][PWM_CHANNELS+1];			// alloc 3 buffers

// pointer to buffers
static pwm_timer_t *buffer1 = &(buffer[0][0]);
static pwm_timer_t *buffer2 = &(buffer[1][0]);
static pwm_timer_t *buffer3 = &(buffer[2][0]);


volatile pwm_timer_t *outputBuffer = &(buffer[0][0]);
static pwm_timer_t *nextBuffer = &(buffer[1][0]);
static pwm_timer_t *initBuffer = &(buffer[2][0]);

void pwmInit() {
	// init ports
	PWM_DDR1 |= PWM_MASK1;
	
	// init timer 1
	TCCR1A = 0;
	TCCR1B = (1<<WGM12);				// ctc with OCR1A as top
	OCR1A = PWM_TOP;
	TIMSK1 |= (1<<OCIE1A) | (1<<OCIE1B);
	OCR1B = 0xffff;						// never reached
}

void pwmSetDutyCycle(pwm_channel_t c, uint16_t _dutyCycle) {
	dutyCycle[c] = _dutyCycle & PWM_TOP;
}

uint16_t pwmGetDutyCycle(pwm_channel_t c) {
	return dutyCycle[c];
}

void pwmStart() {
	TCCR1B |= (1<<CS11) | (1<<CS10);		// 64 prescaler
	//TCCR1B |= (1<<CS12) | (1<<CS10);		// 1024 prescaler
}

void pwmStop() {
	TCCR1B &= ~((1<<CS11) | (1<<CS10) | (1<<CS12));
	PWM_PORT1 &= ~PWM_MASK1; 
}

void pwmUpdate() {
	uint8_t iChannel, iTimer;
	uint8_t lastTimer = 0;
	uint8_t sreg;
	pwm_timer_t *outputbufferTemp;
	
	// select init buffer
	sreg = SREG;
	cli();
	outputbufferTemp = (pwm_timer_t*)outputBuffer;
	SREG = sreg;
	
	if (buffer1 != nextBuffer && buffer2 <= outputbufferTemp) {
		initBuffer = buffer1;
	}
	else if (buffer3 != nextBuffer && buffer3 > outputbufferTemp) {
		initBuffer = buffer3;
	}
	else {
		initBuffer = buffer2;
	}
	
	// init start values
	initBuffer[PWM_START].port1 = (PWM_PORT1 & ~PWM_MASK1);
	initBuffer[PWM_START].nextTime = 0;
	
	// init times
	for (iChannel = 0; iChannel<PWM_CHANNELS; iChannel++) {
		if (dutyCycle[iChannel] != 0) {							// dutyCycle > 0
			initBuffer[PWM_START].port1 |= (1<<(iChannel+PIN_OFFSET));
			
			if (dutyCycle[iChannel] != PWM_TOP) {
				// order time values
				for (iTimer=0; iTimer <lastTimer; iTimer ++) {
					if (dutyCycle[iChannel] == initBuffer[iTimer].nextTime) {	// same dutyCycle
						// set ports
						initBuffer[iTimer+1].port1 &= ~(1<<(iChannel+PIN_OFFSET));
						break;
					}
					if (dutyCycle[iChannel] < initBuffer[iTimer].nextTime) {	// smaller dutyCycle
						// move greater timers
						memmove(&(initBuffer[iTimer+1]), &(initBuffer[iTimer]), (lastTimer-iTimer+1)*sizeof(pwm_timer_t));
						//initBuffer[iTimer] = initBuffer[iTimer+1];
						//initBuffer[iTimer+1] = initBuffer[iTimer+2];
						
						// insert new timer at iTimer
						initBuffer[iTimer].nextTime = dutyCycle[iChannel];
						initBuffer[iTimer+1].port1 = ~(1<<(iChannel+PIN_OFFSET));
						lastTimer++;
						break;
					}
				}
				
				if (iTimer == lastTimer) {
					// insert new timer at iTimer
					initBuffer[iTimer].nextTime = dutyCycle[iChannel];
					initBuffer[iTimer+1].port1 = ~(1<<(iChannel+PIN_OFFSET));
					lastTimer++;
				}
			}
		} 
	}
	
	initBuffer[lastTimer].nextTime = INF;
	initBuffer[PWM_START].nextTime = (initBuffer[PWM_START].nextTime<<8) | (initBuffer[PWM_START].nextTime>>8);
	
	for (iTimer=1; iTimer<=lastTimer; iTimer++) {
		// apply bitmasks
		initBuffer[iTimer].port1 = initBuffer[iTimer-1].port1 & initBuffer[iTimer].port1;
		
		// swap time bytes
		initBuffer[iTimer].nextTime = (initBuffer[iTimer].nextTime<<8) | (initBuffer[iTimer].nextTime>>8);
	}

	//initBuffer[iTimer].port1 = initBuffer[iTimer].port1 << PIN_OFFSET;

	sreg = SREG;
	cli();
	nextBuffer = initBuffer;
	SREG = sreg;
}

// end of pwm cycle
// set GPIOR0 register as periodic task flag
ISR (TIMER1_COMPA_vect, ISR_NAKED) {
	asm(	"push r30"					"\n\t"
			"push r31"					"\n\t"
			"push r24"					"\n\t"
			"lds r30,(%4)"				"\n\t"
			"lds r31,(%4)+1"			"\n\t"
			"ld r24,Z+"					"\n\t"
			"out %0,r24"				"\n\t"
			"ld r24,Z+"					"\n\t"
			"sts %1,r24"				"\n\t"
			"ld r24,Z+"					"\n\t"
			"sts %2,r24"				"\n\t"
			"sts (%5),r30"				"\n\t"
			"sts (%5)+1,r31"			"\n\t"
			"pop r24"					"\n\t"
			"pop r31"					"\n\t"
			"pop r30"					"\n\t"
	//		"sbi %3, 1"					"\n\t"
			"reti"						"\n\t"
	::"M" (_SFR_IO_ADDR(PWM_PORT1)), "M" (_SFR_IO_ADDR(OCR1BH)+0x20), "M" (_SFR_IO_ADDR(OCR1BL)+0x20), "M" (_SFR_IO_ADDR(newCycle)), "m" (nextBuffer), "m" (outputBuffer));
}

// load times
ISR (TIMER1_COMPB_vect, ISR_NAKED) {
	asm(	"push r30"					"\n\t"
			"push r31"					"\n\t"
			"push r24"					"\n\t"
			"lds r30,(%3)"				"\n\t"
			"lds r31,(%3)+1"			"\n\t"
			"ld r24,Z+"					"\n\t"
			"out %0,r24"				"\n\t"
			"ld r24,Z+"					"\n\t"
			"sts %1,r24"				"\n\t"
			"ld r24,Z+"					"\n\t"
			"sts %2,r24"				"\n\t"
			"sts (%3),r30"				"\n\t"
			"sts (%3)+1,r31"			"\n\t"
			"pop r24"					"\n\t"
			"pop r31"					"\n\t"
			"pop r30"					"\n\t"
			"reti"						"\n\t"
	::"M" (_SFR_IO_ADDR(PWM_PORT1)), "M" (_SFR_IO_ADDR(OCR1BH)+0x20), "M" (_SFR_IO_ADDR(OCR1BL)+0x20), "m" (outputBuffer));
}

